/*
 * NAME
 *   mockd - mock daemon for genrc testsuite
 *
 * SYNOPSIS
 *   mockd [-afg] [-i ID] [-l FILE] [-p FILE] [-t N]
 *
 *   mockd -w -p FILE
 *
 * DESCRIPTION
 *   This program mimics the behavior of an average UNIX daemon:
 *   it disconnects from the controlling terminal and "runs" (actually,
 *   waits) in background.  It exits upon delivery of a signal or when
 *   a timeout expires, whichever happens first.
 *
 * OPTIONS:
 *
 *   -a      Append to the logfile (see -l below).
 *   -f      Remain in foreground.
 *   -g      Do an extra fork and start a "grandchild" process.
 *   -i ID   Set program IP, i.e. a string used to identify it in log
 *           messages, instead of the program name.
 *   -l FILE
 *           Write diagnostics to FILE.  If the file exists, it will
 *           be truncated, unless the -a option is specified.
 *   -p FILE
 *           Write PID of the process to FILE.
 *   -t SECONDS
 *           Set timeout.  Upon its expiration, the program will terminate.
 *           The default it 10 seconds.
 *   -w      Wait for the pidfile (supplied with the -p option) to appear
 *           on disk and for the PID (a decimal number followed by a newline
 *           character) to appear in it and exit.  This is used in test
 *           scripts to synchronize processes.
 *
 * SIGNALS
 *   The program reacts on the following signals:
 *
 *   SIGHUP, SIGUSR1
 *      Print a diagnostic message and continue.
 *   SIGALRM, SIGINT, SIGQUIT, SIGUSR2
 *      Print a diagnostic message and terminate.
 *
 *   Delivery of a signal not listed above, causes the program to terminate
 *   abnormally.
 *
 * LICENSE
 *   Copyright (C) 2025 Sergey Poznyakoff
 *
 *   Genrc is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by the
 *   Free Software Foundation; either version 3 of the License, or (at your
 *   option) any later version.
 *
 *   Genrc is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with genrc. If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <sys/wait.h>

char *progname;
int foreground;
int timeout = 10;
char *logfile;
int logappend;
char *pidfile;

static void
rmpidfile(void)
{
	unlink(pidfile);
}

static void
signul(int sig)
{
	/* nothing */
}

static int sigtab[] = {
	SIGALRM,
	SIGHUP,
	SIGINT,
	SIGQUIT,
	SIGTERM,
	SIGUSR1,
	SIGUSR2,
	0
};

static char *signame[] = {
	"SIGALRM",
	"SIGHUP",
	"SIGINT",
	"SIGQUIT",
	"SIGTERM",
	"SIGUSR1",
	"SIGUSR2",
	NULL
};

static char const *
getsigname(int signo)
{
	int i;

	for (i = 0; sigtab[i]; i++)
		if (sigtab[i] == signo)
			return signame[i];
	return NULL;
}

/* Wait until pidfile is created */
static void
waitpidfile(pid_t pid)
{
	int fd;
	long n = 0;
	char c;
	int rb;

	alarm(timeout);
	while ((fd = open(pidfile, O_RDONLY)) == -1) {
		if (errno != ENOENT) {
			perror("< pidfile");
			abort();
		}
				}
	while ((rb = read(fd, &c, 1)) >= 0) {
		if (rb == 1) {
			if (c == '\n')
				break;
			assert('0' <= c && c <= '9');
			n = n * 10 + c - '0';
		}
	}
	close(fd);
	if (rb == -1) {
		perror("read pidfile");
		abort();
	}
	if (pid != (pid_t)-1)
		assert(n == pid);
}

static void
wrapper(void)
{
	int status;
	pid_t pid;

	pid = fork();
	switch (pid) {
	case -1:
		perror("fork");
		abort();
	case 0:
		break;
	default:
		if (waitpid(pid, &status, 0) == -1) {
			perror("waitpid");
			_exit(1);
		}
		if (WIFSIGNALED(status)) {
			fprintf(stderr, "%s: child exited on signal %d\n",
				progname, WTERMSIG(status));
			_exit(1);
		}
		if (WIFEXITED(status))
			_exit(WEXITSTATUS(status));
		fprintf(stderr, "%s: child exit status %d\n",
			progname, status);
		_exit(1);
	}
}

int
main(int argc, char **argv)
{
	int i;
	pid_t pid;
	sigset_t sigs;
	int signo;
	struct sigaction act;
	char const *reason = NULL;
	int waitpid_opt = 0;
	int group_opt = 0;

	if ((progname = strrchr(argv[0], '/')) == NULL)
		progname = argv[0];
	else
		progname++;

	while ((i = getopt(argc, argv, "afgi:l:p:t:w")) != EOF) {
		switch (i) {
		case 'a':
			logappend = 1;
			break;

		case 'f':
			foreground = 1;
			break;

		case 'g':
			group_opt = 1;
			break;

		case 'i':
			progname = optarg;
			break;

		case 'l':
			logfile = optarg;
			break;

		case 'p':
			pidfile = optarg;
			break;

		case 't':
			timeout = atoi(optarg);
			assert(timeout > 0);
			break;

		case 'w':
			waitpid_opt = 1;
			break;

		default:
			return 1;
		}
	}

	if (waitpid_opt) {
		assert(pidfile != 0);
		waitpidfile((pid_t)-1);
		return 0;
	}

	if (!foreground) {
		int fd;

		pid = fork();
		if (pid == -1) {
			perror("fork");
			abort();
		}
		if (pid) {
			if (pidfile)
				waitpidfile(group_opt ? (pid_t)-1 : pid);
			return 0;
		}
		if (group_opt)
			wrapper();
		close(0);
		close(1);
		close(2);
		if (open("/dev/null", O_RDONLY) != 0)
			exit(125);
		if (logfile)
			fd = open(logfile,
				  O_CREAT|O_WRONLY|(logappend ? 0 : O_TRUNC),
				  0600);
		else
			fd = open("/dev/null", O_WRONLY);
		if (logappend)
			assert(lseek(fd, 0, SEEK_END) != -1);
		if (fd != 1)
			exit(126);
		if (dup(1) != 2)
			exit(127);
	} else if (group_opt)
		wrapper();

	setvbuf(stdout, NULL, _IOLBF, 0);
	setvbuf(stderr, NULL, _IOLBF, 0);

	pid = getpid();
	printf("%s: started\n", progname);
	if (pidfile) {
		FILE *fp = fopen(pidfile, "w");
		if (!fp) {
			perror(pidfile);
			abort();
		}
		fprintf(fp, "%ld\n", (long) pid);
		fclose(fp);
		atexit(rmpidfile);
	}

	alarm(timeout);
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	act.sa_handler = signul;

	sigemptyset(&sigs);
	for (i = 0; sigtab[i]; i++) {
		sigaddset(&sigs, sigtab[i]);
		sigaction(sigtab[i], &act, NULL);
	}

	while (reason == NULL) {
		if (sigwait(&sigs, &signo) < 0) {
			perror("sigwait");
			abort();
		}
		switch (signo) {
		case SIGALRM:
			reason = "timeout";
			break;

		case SIGHUP:
		case SIGUSR1:
			printf("%s: %s\n", progname, getsigname(signo));
			alarm(timeout);
			break;

		default:
			reason = getsigname(signo);
			assert(reason != NULL);
		}
	}
	printf("%s: terminating on %s\n", progname, reason);
	return 0;
}
